<?php
/**
 * @package    AqurateAi_Recommendations
 * @copyright  Copyright (c) 2023 Clever++
 * @author     Clever++ <office@cleverplusplus.com>
 */

namespace AqurateAi\Recommendations\Helper;

use Magento\Backend\Model\Auth\Session;
use Magento\Cron\Model\Config\Source\Frequency;
use Magento\Framework\App\Cache\TypeListInterface;
use Magento\Framework\App\Config\ScopeConfigInterface;
use Magento\Framework\App\Config\Value;
use Magento\Framework\App\Config\ValueFactory;
use Magento\Framework\App\RequestInterface;
use Magento\Framework\Data\Collection\AbstractDb;
use Magento\Framework\Message\ManagerInterface;
use Magento\Framework\Model\Context;
use Magento\Framework\Model\ResourceModel\AbstractResource;
use Magento\Framework\Registry;
use Magento\Integration\Api\IntegrationServiceInterface;
use Magento\Integration\Api\OauthServiceInterface as IntegrationOauthService;
use Magento\Integration\Model\AuthorizationService;
use Magento\Integration\Model\IntegrationFactory;
use Magento\Store\Model\StoreManagerInterface;

class CronConfig extends Value
{
    const CROSS_SELL_CRON_STRING_PATH = 'crontab/default/jobs/aqurateai_recommendations_crosssell/schedule/cron_expr';
    const CROSS_SELL_CRON_MODEL_PATH = 'crontab/default/jobs/aqurateai_recommendations_cron_job/run/cross_sell_model';
    const SUBSTITUTES_CRON_STRING_PATH = 'crontab/default/jobs/aqurateai_recommendations_substitutes/schedule/cron_expr';
    const SUBSTITUTES_CRON_MODEL_PATH = 'crontab/default/jobs/aqurateai_recommendations_cron_job/run/substitutes_model';
    const INTEGRATION_NAME = 'aqurate';
    const INTEGRATION_CALLBACK_URL = 'https://api.aqurate.ai/v1/integrations/magento-2-callback';
    const INTEGRATION_IDENTIFY_LINK_URL = 'https://app.aqurate.ai/login';

    /**
     * Integration service
     *
     * @var \Magento\Integration\Api\IntegrationServiceInterface
     */
    protected $_integrationService;

    /**
     * @var IntegrationFactory
     */
    protected $_integrationFactory;

    /**
     * @var IntegrationOauthService
     */
    protected $_oauthService;

    /**
     * @var AuthorizationService
     */
    protected $_authorizationService;

    /**
     * @var ValueFactory
     */
    protected $_configValueFactory;

    /**
     * @var mixed|string
     */
    protected $_runModelPath = '';

    /**
     * @var ScopeInterface
     */
    protected $_scopeConfig;

    /**
     * @var StoreManagerInterface
     */
    protected $_storeManager;

    /**
     * @var RequestInterface
     */
    protected $_request;

    /**
     * @var Session
     */
    protected $_session;

    /**
     * @var ManagerInterface
     */
    private $_messageManager;

    /**
     * @param Context                     $context
     * @param Registry                    $registry
     * @param ScopeConfigInterface        $config
     * @param TypeListInterface           $cacheTypeList
     * @param ValueFactory                $configValueFactory
     * @param IntegrationServiceInterface $integrationService
     * @param IntegrationFactory          $integrationFactory
     * @param IntegrationOauthService     $oauthService
     * @param AuthorizationService        $authorizationService
     * @param StoreManagerInterface       $storeManager
     * @param RequestInterface            $request
     * @param Session                     $session
     * @param ManagerInterface            $messageManager
     * @param AbstractResource|null       $resource
     * @param AbstractDb|null             $resourceCollection
     * @param string                      $runModelPath
     * @param array                       $data
     */
    public function __construct(
        Context                     $context,
        Registry                    $registry,
        ScopeConfigInterface        $config,
        TypeListInterface           $cacheTypeList,
        ValueFactory                $configValueFactory,
        IntegrationServiceInterface $integrationService,
        IntegrationFactory          $integrationFactory,
        IntegrationOauthService     $oauthService,
        AuthorizationService        $authorizationService,
        StoreManagerInterface       $storeManager,
        RequestInterface            $request,
        Session                     $session,
        ManagerInterface            $messageManager,
        AbstractResource            $resource = null,
        AbstractDb                  $resourceCollection = null,
        string                      $runModelPath = '',
        array                       $data = []
    ) {
        $this->_runModelPath         = $runModelPath;
        $this->_configValueFactory   = $configValueFactory;
        $this->_integrationService   = $integrationService;
        $this->_integrationFactory   = $integrationFactory;
        $this->_oauthService         = $oauthService;
        $this->_authorizationService = $authorizationService;
        $this->_scopeConfig          = $config;
        $this->_storeManager         = $storeManager;
        $this->_request              = $request;
        $this->_session              = $session;
        $this->_messageManager       = $messageManager;

        parent::__construct(
            $context,
            $registry,
            $config,
            $cacheTypeList,
            $resource,
            $resourceCollection,
            $data
        );
    }

    /**
     * @return CronConfig
     * @throws \Exception
     */
    public function afterSave(): CronConfig
    {
        $crossSellTime      = $this->getData('groups/advanced/fields/cross_sell_time/value');
        $crossSellFrequency = $this->getData('groups/advanced/fields/cross_sell_frequency/value');

        $substitutesTime      = $this->getData('groups/advanced/fields/substitutes_time/value');
        $substitutesFrequency = $this->getData('groups/advanced/fields/substitutes_frequency/value');

        $crossSellCronExprArray = [
            intval($crossSellTime[1]),
            intval($crossSellTime[0]),
            $crossSellFrequency == Frequency::CRON_MONTHLY ? '1' : '*',
            '*',
            $crossSellFrequency == Frequency::CRON_WEEKLY ? '1' : '*'
        ];

        $substitutesSellCronExprArray = [
            intval($substitutesTime[1]),
            intval($substitutesTime[0]),
            $substitutesFrequency == Frequency::CRON_MONTHLY ? '1' : '*',
            '*',
            $substitutesFrequency == Frequency::CRON_WEEKLY ? '1' : '*'
        ];

        $crossSellCronExprString       = join(' ', $crossSellCronExprArray);
        $substitutesSellCronExprString = join(' ', $substitutesSellCronExprArray);

        try {

            /**
             * Cross-Sells
             */
            $this->_configValueFactory->create()->load(self::CROSS_SELL_CRON_STRING_PATH, 'path')
                ->setValue($crossSellCronExprString)->setPath(self::CROSS_SELL_CRON_STRING_PATH)->save();

            $this->_configValueFactory->create()->load(self::CROSS_SELL_CRON_MODEL_PATH, 'path')
                ->setValue($this->_runModelPath)->setPath(self::CROSS_SELL_CRON_MODEL_PATH)->save();

            /**
             * Substitutes
             */
            $this->_configValueFactory->create()->load(self::SUBSTITUTES_CRON_STRING_PATH, 'path')
                ->setValue($substitutesSellCronExprString)->setPath(self::SUBSTITUTES_CRON_STRING_PATH)->save();

            $this->_configValueFactory->create()->load(self::SUBSTITUTES_CRON_MODEL_PATH, 'path')
                ->setValue($this->_runModelPath)->setPath(self::SUBSTITUTES_CRON_MODEL_PATH)->save();
        } catch (\Exception $e) {
            $this->_logger->error($e->getMessage());
            throw new \Exception(__('We can\'t save the cron expression.'));
        }

        return parent::afterSave();
    }

    /**
     * @param $integrationData
     * @return CronConfig|\Magento\Integration\Model\Integration
     */
    public function getIntegration()
    {
        $email = $this->_session->getUser()->getEmail();

        $integrationData = [
            'name'              => self::INTEGRATION_NAME,
            'email'             => $email,
            'status'            => 1,
            'endpoint'          => self::INTEGRATION_CALLBACK_URL,
            'identity_link_url' => self::INTEGRATION_IDENTIFY_LINK_URL
        ];

        try {
            $int = $this->_integrationService->findByName(self::INTEGRATION_NAME);
            if ($int->getId()) {
                if ($int->getEmail() == $email) {
                    return $int;
                } else {
                    $integration = $this->_integrationFactory->create()->load($int->getId());
                }
            } else {
                $integration = $this->_integrationFactory->create()->setData($integrationData);
            }

            return $integration;
        } catch (\Exception $e) {
            $this->_logger->error($e->getMessage());

            return null;
        }
    }

    /**
     * @param $integration
     * @return \Magento\Integration\Model\Oauth\Consumer|null
     */
    public function getConsumer($integration)
    {
        try {
            /**
             * create consumer
             */
            $consumer = $this->_oauthService->loadConsumer($integration->getConsumerId());

            if ($consumer->getId()) {
                return $consumer;
            }

            return $this->_oauthService->createConsumer(['name' => self::INTEGRATION_NAME]);
        } catch (\Exception $e) {
            $this->_logger->error($e->getMessage());

            return null;
        }
    }

    /**
     * @return string
     * @throws \Magento\Framework\Exception\LocalizedException
     * @throws \Magento\Framework\Exception\NoSuchEntityException
     * @throws \Magento\Framework\Oauth\Exception
     */
    public function getAuthUrl()
    {
        $consumerKey = $this->getConsumerKey();

        if ($consumerKey) {
            $baseUrl  = $this->_storeManager->getStore()->getBaseUrl(\Magento\Framework\UrlInterface::URL_TYPE_WEB) .
                "admin/admin/integration/loginSuccessCallback/key/" . $this->_request->getParam('key');

            $url = self::INTEGRATION_IDENTIFY_LINK_URL . "?oauth_consumer_key=" . $consumerKey;
            $url .= '&success_call_back =' . urlencode($baseUrl);

            return $url;
        }

        return null;
    }

    /**
     * @return string
     */
    public function getIdentityUrl()
    {
        return self::INTEGRATION_IDENTIFY_LINK_URL;
    }

    /**
     * @return array|mixed|string|void|null
     * @throws \Magento\Framework\Exception\LocalizedException
     * @throws \Magento\Framework\Oauth\Exception
     */
    public function getConsumerKey()
    {
        $integration = $this->getIntegration();

        if (is_object($integration) && $integration->getId()) {
            $consumer = $this->getConsumer($integration);

            if (is_object($consumer) && $consumer->getId()) {
                $consumer = $this->_oauthService->loadConsumer($consumer->getId());

                return $consumer->getKey();
            }
        }

        return null;
    }

    /**
     * @return void
     */
    public function createIntegration()
    {
        try {
            /**
             * save integration
             */
            $integration = $this->getIntegration();

            $consumer = $this->getConsumer($integration);

            if (is_object($consumer) && $consumer->getId()) {

                /**
                 * save consumer id to integration
                 */
                $integration->setConsumerId($consumer->getId());

                $integration->save();

                $this->_oauthService->createAccessToken($consumer->getId());

                /**
                 * grant permissions
                 */
                $permissions = [
                    'Magento_Sales::sales',
                    'Magento_Sales::sales_operation',
                    'Magento_Sales::sales_order',
                    'Magento_Sales::actions',
                    'Magento_Sales::actions_view',
                    'Magento_Catalog::catalog',
                    'Magento_Catalog::catalog_inventory',
                    'Magento_Catalog::products',
                    'Magento_Catalog::categories',
                    'Magento_Customer::customer',
                    'Magento_Customer::manage',
                    'Magento_Customer::actions'
                ];

                $this->_authorizationService->grantPermissions($integration->getId(), $permissions);
                $this->_messageManager->addSuccessMessage(__('The integration has been created. Please authorize it.'));
                $this->_logger->info(__('The integration has been created.'));
            }
        } catch (\Exception $e) {
            $this->_messageManager->addErrorMessage(__($e->getMessage()));
            $this->_logger->error($e->getMessage());
        }
    }
}
