<?php
/**
 * @package    AqurateAi_Recommendations
 * @copyright  Copyright (c) 2023 Clever++
 * @author     Clever++ <office@cleverplusplus.com>
 */

declare(strict_types=1);

namespace AqurateAi\Recommendations\Model;

use AqurateAi\Recommendations\Api\Data\CrossSellInterface;
use AqurateAi\Recommendations\Api\Data\SubstitutesInterface;
use AqurateAi\Recommendations\Model\Cache\Type\CacheTypeRelated;
use Magento\Catalog\Api\ProductRepositoryInterface;
use Magento\Catalog\Model\Product\Attribute\Source\Status;
use Magento\Catalog\Model\Product\Visibility;
use Magento\Catalog\Model\Product\Visibility as ProductVisibility;
use Magento\Catalog\Model\ResourceModel\Product\CollectionFactory;
use Magento\Framework\Api\SearchCriteriaBuilder;
use Magento\Framework\App\CacheInterface;
use Magento\Framework\App\Config\ScopeConfigInterface;
use Magento\Framework\Exception\LocalizedException;
use Magento\Framework\Model\AbstractModel;
use Magento\Framework\Serialize\SerializerInterface;
use Magento\Store\Model\ScopeInterface;
use Psr\Log\LoggerInterface;

class DataProcessor extends AbstractModel
{
    const TYPE_SUBSTITUTES = 'substitutes';
    const TYPE_CROSS_SELL = 'cross-sell';
    const TYPE_CROSS_SELL_CART = 'cross-sell-cart';

    /**
     * @var CacheInterface
     */
    protected $_cache;

    /**
     * @var ScopeConfigInterface
     */
    protected $_scopeConfig;

    /**
     * Catalog product visibility
     *
     * @var ProductVisibility
     */
    protected $_catalogProductVisibility;

    /**
     * @var ProductRepositoryInterface
     */
    protected $_productRepository;

    /**
     * @var SerializerInterface
     */
    protected $_serializer;

    /**
     * @var \AqurateAi\Recommendations\Model\CrossSellRepository
     */
    protected $_crossSellRepository;

    /**
     * @var \AqurateAi\Recommendations\Model\SubstitutesRepository
     */
    protected $_substitutesRepository;

    /**
     * @var SearchCriteriaBuilder
     */
    protected $_searchCriteriaBuilder;

    /**
     * @var \Magento\Catalog\Model\ResourceModel\Product\CollectionFactory
     */
    protected $_productCollectionFactory;

    /**
     * @var \Magento\Catalog\Model\Product\Attribute\Source\Status
     */
    protected $_productStatus;

    /**
     * @var \Magento\Catalog\Model\Product\Visibility
     */
    protected $_productVisibility;

    /**
     * @var CrossSellInterface
     */
    protected $_crossSellInterface;

    /**
     * @var SubstitutesInterface
     */
    protected $_substitutesInterface;

    /**
     * @var LoggerInterface
     */
    protected $_logger;

    /**
     * @var mixed
     */
    protected $apiUrl;

    /**
     * @var mixed
     */
    protected $apiEmail;

    /**
     * @var mixed
     */
    protected $apiPassword;

    /**
     * @var
     */
    protected $_storeManager;

    /**
     * @param \Magento\Framework\Model\Context                             $context
     * @param \Magento\Framework\Registry                                  $registry
     * @param \Magento\Framework\Model\ResourceModel\AbstractResource|null $resource
     * @param CacheInterface                                               $cache
     * @param ScopeConfigInterface                                         $scopeConfig
     * @param Visibility                                                   $catalogProductVisibility
     * @param SerializerInterface                                          $serializer
     * @param ProductRepositoryInterface                                   $productRepository
     * @param \AqurateAi\Recommendations\Model\CrossSellRepository         $crossSellRepository
     * @param CrossSellInterface                                           $crossSellInterface
     * @param \AqurateAi\Recommendations\Model\SubstitutesRepository       $substitutesRepository
     * @param SubstitutesInterface                                         $substitutesInterface
     * @param SearchCriteriaBuilder                                        $searchCriteriaBuilder
     * @param CollectionFactory                                            $collectionFactory
     * @param Status                                                       $status
     * @param Visibility                                                   $visibility
     * @param LoggerInterface                                              $logger
     * @param \Magento\Framework\Data\Collection\AbstractDb|null           $resourceCollection
     * @param array                                                        $data
     */
    public function __construct(
        \Magento\Framework\Model\Context                        $context,
        \Magento\Framework\Registry                             $registry,
        \Magento\Framework\Model\ResourceModel\AbstractResource $resource = null,
        CacheInterface                                          $cache,
        ScopeConfigInterface                                    $scopeConfig,
        ProductVisibility                                       $catalogProductVisibility,
        SerializerInterface                                     $serializer,
        ProductRepositoryInterface                              $productRepository,
        CrossSellRepository                                     $crossSellRepository,
        CrossSellInterface                                      $crossSellInterface,
        SubstitutesRepository                                   $substitutesRepository,
        SubstitutesInterface                                    $substitutesInterface,
        SearchCriteriaBuilder                                   $searchCriteriaBuilder,
        CollectionFactory                                       $collectionFactory,
        Status                                                  $status,
        Visibility                                              $visibility,
        LoggerInterface                                         $logger,
        \Magento\Framework\Data\Collection\AbstractDb           $resourceCollection = null,
        array                                                   $data = []
    ) {
        $this->_cache                    = $cache;
        $this->_scopeConfig              = $scopeConfig;
        $this->_catalogProductVisibility = $catalogProductVisibility;
        $this->_serializer               = $serializer;
        $this->_productRepository        = $productRepository;
        $this->_crossSellRepository      = $crossSellRepository;
        $this->_crossSellInterface       = $crossSellInterface;
        $this->_substitutesRepository    = $substitutesRepository;
        $this->_substitutesInterface     = $substitutesInterface;
        $this->_searchCriteriaBuilder    = $searchCriteriaBuilder;
        $this->_productCollectionFactory = $collectionFactory;
        $this->_productStatus            = $status;
        $this->_productVisibility        = $visibility;
        $this->_logger                   = $logger;

        $this->apiKey   = $this->_scopeConfig->getValue('recommendations/general/api_key', ScopeInterface::SCOPE_STORE);
        $this->apiUrl   = $this->_scopeConfig->getValue('recommendations/advanced/api_url', ScopeInterface::SCOPE_STORE);
        $this->apiLimit = $this->_scopeConfig->getValue('recommendations/advanced/api_limit', ScopeInterface::SCOPE_STORE);

        parent::__construct(
            $context,
            $registry,
            $resource,
            $resourceCollection,
            $data
        );
    }

    /**
     * @param $pid
     * @param $type
     * @return array|null
     */
    public function apiGetProducts($pid, $type, $no = 10, $limit = 1000)
    {
        if (!$this->apiKey) {
            return null;
        }

        $ch = curl_init();
        $url  = $this->apiUrl . 'recs/item/' . $type . '?product_id=' . $pid . '&top=' . $this->apiLimit . '&recs_no=' . $no . '&limit=' . $limit;

        curl_setopt($ch, CURLOPT_URL, $url);
        curl_setopt($ch, CURLOPT_RETURNTRANSFER, 1);
        curl_setopt($ch, CURLOPT_CUSTOMREQUEST, 'GET');

        $headers   = [];
        $headers[] = 'Accept: application/json';
        $headers[] = 'X-Auth-Token: ' . $this->apiKey;
        curl_setopt($ch, CURLOPT_HTTPHEADER, $headers);

        $result = curl_exec($ch);
        if (curl_errno($ch)) {
            echo 'Error:' . curl_error($ch);
        }
        curl_close($ch);

        $result = json_decode($result);

        if (isset($result->recs)) {
            return $result->recs;
        }
        return [];
    }

    /**
     * @param $pid
     * @param $type
     * @return array|false|string[]
     * @throws LocalizedException
     */
    public function dbGetProducts($pid, $type)
    {
        switch ($type) {
            case self::TYPE_SUBSTITUTES:
                $repository = $this->_substitutesRepository;
                break;
            default:
                $repository = $this->_crossSellRepository;
                break;
        }

        $collection = $repository->getList(
            $this->_searchCriteriaBuilder->addFilter('product_id', $pid)
                ->create()
        );

        $result = [];
        if ($collection->getTotalCount()) {
            $collection = $collection->getItems();
            foreach ($collection as $col) {
                $result = explode(',', $col->getRelatedIds());
            }
        }
        return $result;
    }

    /**
     * @param $cPid
     * @param $type
     * @return array|null
     * @throws LocalizedException
     */
    public function getItems($cPid, $type): ?array
    {
        $itemCollectionRelated = null;
        $cacheKey              = CacheTypeRelated::TYPE_IDENTIFIER_AQURATE . '_' . $type . '_' . $cPid;
        $cacheTag              = CacheTypeRelated::CACHE_TAG_AQURATE . '_' . $type . '_' . $cPid;

        $cachedData = $this->_cache->load($cacheKey);

        $productList = null;
        if ($cachedData) {
            $productList = $this->_serializer->unserialize($cachedData);
        }

        if (null == $productList) {
            $productList = $this->dbGetProducts($cPid, $type);

            if (!is_array($productList)) {
                return null;
            }

            $this->_cache->save(
                $this->_serializer->serialize($productList),
                $cacheKey,
                [$cacheTag],
                86400
            );
        }

        foreach ($productList as $id) {
            try {
                $productItem        = $this->_productRepository->getById($id);
                $product_status     = $productItem->getData('status');
                $product_visibility = $productItem->getData('visibility');

                if ($product_status == 1 && $product_visibility != 1) {
                    $itemCollectionRelated[] = $productItem;
                }
            } catch (\Exception $e) {
                continue;
            }
        }
        return $itemCollectionRelated;
    }

    /**
     * @return void|null
     * @throws LocalizedException
     */
    public function ingestRecommendations($type, $products = null, $no = 10, $limit = 1000)
    {
        if (null == $products) {
            $products = $this->getVisibleProducts();
        }

        $entityInterface = $entityRepository = null;
        foreach ($products as $productId) {
            $related = $this->apiGetProducts($productId, $type, $no, $limit);

            if ($related) {
                $related = (array)$related;
                $related = array_column($related, 'rec_product_id');
                $related = implode(',', $related);
            } else {
                continue;
            }

            switch ($type) {
                case self::TYPE_CROSS_SELL:
                    $entityInterface  = $this->_crossSellInterface;
                    $entityRepository = $this->_crossSellRepository;
                    break;

                case self::TYPE_SUBSTITUTES:
                    $entityInterface  = $this->_substitutesInterface;
                    $entityRepository = $this->_substitutesRepository;
                    break;
            }

            if ($entityInterface && $entityRepository) {
                $this->saveRelatedInfo($productId, $related, $entityInterface, $entityRepository, $type);
            }
        }
    }

    /**
     * @param $productId
     * @param $related
     * @param $entityRepository
     * @param $entityInterface
     * @param $type
     * @return void|null
     */
    protected function saveRelatedInfo($productId, $related, $entityInterface, $entityRepository, $type)
    {
        $repository = $entityRepository->getList(
            $this->_searchCriteriaBuilder->addFilter('product_id', $productId)
                ->create()
        );

        $id = null;
        $entityInterface->setEntityId($id);
        if ($repository->getTotalCount() > 0) {
            $items = $repository->getItems();
            foreach ($items as $item) {
                $id = $item->getentityId();
                break;
            }

            if (null == $id) {
                return null;
            }

            $entityInterface->setEntityId($id);
        }

        $entityInterface->setProductId($productId);
        $entityInterface->setRelatedIds($related);
        $entityInterface->setUpdatedAt(date('Y-m-d H:i:s'));

        $logMsd = 'Aqurate.ai ' . $type . ': Magento Product ID: ' . $productId;

        try {
            $entityRepository->save($entityInterface);
            unset($entityRepository);
            $this->_logger->info($logMsd . ' | Related: ' . $related);
            echo $logMsd . PHP_EOL;
        } catch (\Exception $e) {
            $this->_logger->error($logMsd . ' | Related: ' . $related);
            $this->_logger->error($e->getMessage());
        }
    }

    /**
     * @return array|mixed|null
     * @throws LocalizedException
     */
    protected function getVisibleProducts()
    {
        $collection = $this->_productCollectionFactory->create();
        $collection->addAttributeToSelect('entity_id');
        $collection->joinAttribute('status', 'catalog_product/status', 'entity_id', null, 'inner');
        $collection->joinAttribute('visibility', 'catalog_product/visibility', 'entity_id', null, 'inner');
        $collection->addAttributeToFilter('status', ['in' => $this->_productStatus->getVisibleStatusIds()])
            ->addAttributeToFilter('visibility', ['in' => $this->_productVisibility->getVisibleInSiteIds()]);

        $products = [];

        foreach ($collection->getItems() as $item) {
            $products[] = $item->getData('entity_id');
        }

        return $products;
    }
}
